#ifndef __VRML_H_
#define __VRML_H_

#include "vrmlgen.h"
#define space " "
#define TRUE 1
#define FALSE 0


class SFBool {
 private:
  bool boolvalue;
 public:
  inline void insert(bool invalue)
    {
      boolvalue = invalue;
    }
  inline void print()
    {
      if(boolvalue == TRUE)
	cout << "TRUE\n ";
      else
	cout << "FALSE \n";
    }
  
};


class SFColor {

 public:
  float s1,s2,s3;
  

  
  inline void insert(float in1,float in2,float in3)
    {
      s1 = in1;
      s2 = in2;
      s3 = in3;
    }
  
  inline void print()
    {
      printf("%.1f %.1f %.1f\n",s1,s2,s3);
    }
  inline void print_no_eol()
    {
       printf("%.1f %.1f %.1f",s1,s2,s3);
    }

};


class SFFloat {
 private:
  float f1;
 public:
  
  inline void insert(float in1)
    {
      f1 = in1;
    }

  inline void print()
    {
	printf("%1.1f\n",f1);
    }
      
};

class SFRotation {
 private:
  float rlist[4];
 public:
  inline void insert(float r1,float r2,float r3,float r4)
    {
      rlist[0] = r1;rlist[1] = r2;rlist[2]=r3;rlist[3] = r4;
    }
  inline void print()
    {
      cout << rlist[0] << space << rlist[1] << space
	<< rlist[2] << space << rlist[3] << endl;
    }
};

class coord {

 private:
  vector<SFColor> pointlist;
 public:
  inline void insert(SFColor newpoint)
    {
      pointlist.push_back(newpoint);
    }
  inline void print()
    {
      cout << "coord Coordinate {\n";
      cout << "point [\n";
      for(int i=0;i<pointlist.size();i++)
	{
	  pointlist[i].print_no_eol();
	  cout << ",\n";
	}
      cout << "]\n";
      cout << "}\n";
    }
  

};

class coordIndex {

 private:
  vector<int> coord_list;
 public:
  inline void insert(int p1,int p2)
    {
      coord_list.push_back(p1);
      coord_list.push_back(p2);
      coord_list.push_back(-1);
    }
  inline void print()
    {
      cout << "coordIndex [\n";
      for(int i=0;i<coord_list.size();i+=3)
	{
	  for(int j=i;j<i+3;j++)
	  {
	    cout << coord_list[j] << ", ";
	  }
	  cout << endl;
	}
      cout << "]\n";
    }


};


class color {
 private:
  vector<SFColor> colorlist;
 public:
  inline void insert(SFColor newcolor)
    {
      colorlist.push_back(newcolor);
    }
  inline void print()
    {
      cout << "color Color {\n";
      cout << "color [\n";
      for(int i=0;i<colorlist.size();i++)
	{
	  colorlist[i].print(); cout << ",\n";
	}
      cout << "]\n";
      cout << "}\n";
    }
};


class IndexedLineSet {
 public:
  coord coord_object;
  coordIndex index_object;
  
  inline void print()
    {
      cout << "IndexedLineSet {\n";
      coord_object.print();
      index_object.print();
      cout << "}\n";
    }


};





class material {

 public:
  SFColor diffuseColor;
  SFColor emissiveColor;
  SFFloat transparency;
  SFFloat ambientIntensity;
  SFColor specularColor;
  SFFloat shininess;

  material()
    {
      diffuseColor.insert(0.8,0.8,0.8);
      emissiveColor.insert(0.0,0.0,0.0);
      transparency.insert(0.0);
      ambientIntensity.insert(0.2);
      specularColor.insert(0.0,0.0,0.0);
      shininess.insert(0.0);
    }

  inline void print()
    {
      cout << "material Material {\n";
      cout << "diffuseColor "; diffuseColor.print();
      cout << "emissiveColor "; emissiveColor.print();
      cout << "transparency "; transparency.print();
      cout << "ambientIntensity "; ambientIntensity.print();
      cout << "specularColor "; specularColor.print();
      cout << "shininess "; shininess.print();
      cout << "}\n";
    }

};


class appearance {

  
 public:
  material material_node;
  
  
  inline void print()
    {
      cout << "appearance Appearance {\n";
      material_node.print();
      cout << "}\n";

    }

};


class box {
 private:
  SFColor f1;
 public:
  box()
    {
      f1.insert(2.0,2.0,2.0);
    }
  

  insert(float in1,float in2,float in3)
    {
      f1.insert(in1,in2,in3);
    }
  
  inline void print()
    {
      cout << "Box {\n";
      cout << "size ";
      f1.print();
      cout << "}\n";
    }
};

class cone {
 public:
  SFFloat bottomRadius;
  SFFloat height;
  SFBool side;
  SFBool bottom;

  cone()
    {
      bottomRadius.insert(1.0);
      height.insert(2.0);
      side.insert(TRUE);
      bottom.insert(TRUE);
    }
  inline void print()
    {
      cout << "Cone {\n";
      cout << "bottomRadius "; bottomRadius.print();
      cout << "height "; height.print();
      cout << "side "; side.print();
      cout << "bottom "; bottom.print();
      cout << "}\n";
    }
};


class text {

 private:
  char string[50];
  int length;
 public:
  text() { length=1;}

  inline void insert(char* entry)
    {
      strcpy(string,entry);
    }

  inline void print()
    {
      cout << "Text {\n";
      cout << "string \"" << string << "\"\n";
      cout << "length " << length << endl;
      cout << "}\n";
    }

};

template <class A,class B, class C,class D>
class geom_container {
  
 private:
  int flag;
  
 public:
  A* a1;
  B* b1;
  C* c1;
  D* d1;

  
  geom_container()
    {
      a1 = NULL;
      b1 = NULL;
      c1 = NULL;
      d1 = NULL;
      flag = -1;
    }
  inline void insert(geom_container<A,B,C,D> newcontainer)
    {
      if(newcontainer.a1 != NULL)
	{
	  a1 = newcontainer.a1;
	  flag = 1;
	}
      else if(newcontainer.b1 != NULL)
	{
	  b1 = newcontainer.b1;
	  flag = 2;
	}
      else if(newcontainer.c1 != NULL)
	{
	  c1 = newcontainer.c1;
	  flag =3;
	}
      else if (newcontainer.d1 != NULL)
	{
	  d1 = newcontainer.d1;
	  flag = 4;
	}
    }
  

  inline void insert(A ina)
    {
      a1 = new A(ina);
      flag = 1;
    }

  inline void insert(B inb)
    {
      b1 = new B(inb);
      flag = 2;
    }

    inline void insert(C inc)
      {
	c1 = new C(inc);
	flag = 3;
      }
  inline void insert(D ind)
    {
      d1 = new D(ind);
      flag = 4;
    }

  inline void print()
    {
      if(flag == 1)
	a1->print();
      else if(flag == 2)
	b1->print();
      else if(flag == 3)
	c1->print();
      else if(flag == 4)
	d1->print();
    }
  

};



class geometry {

 private:
  geom_container<box,cone,text,IndexedLineSet> container;
 public:
  inline void insert(geom_container<box,cone,text,IndexedLineSet>
		     newcontainer)
    {
      container.insert(newcontainer);
    }

  inline void print()
    {
      cout << "geometry ";
      container.print();
    }
 

};



class shape
{
 private:

  geometry geometry_node;
 public:
  appearance appearance_node;
  insert(geom_container<box,cone,text,IndexedLineSet> object)
    {
      geometry_node.insert(object);
    }

  void print()
    {
      cout << "Shape {\n";
      appearance_node.print();
      geometry_node.print();
      cout << "}\n";
      
    }
  

};

class billboard {

 private:
  vector<shape> blist;
 public:
  SFColor axisOfRotation;
  billboard ()
    {
      axisOfRotation.insert(0.0,0.0,0.0);
    }

  inline void insert(shape shape_object)
    {
      blist.push_back(shape_object);
    }
  inline void print()
    {
      cout << "Billboard {\n";
      cout << "axisOfRotation "; axisOfRotation.print();
      cout << "children [\n";
      for(int i=0;i<blist.size();i++)
	{
	  blist[i].print();
	  cout << ",\n";
	}
      cout << "]\n";
      cout << "}\n";
    }

};

class pointlight {

 public:
  SFBool on;
  SFColor location;
  SFFloat radius;
  SFFloat intensity;
  SFFloat ambientIntensity;
SFColor color;
  
  pointlight()
    {
      on.insert(TRUE);
      location.insert(0,0,0);
      radius.insert(10);
      intensity.insert(.7);
      ambientIntensity.insert(0.5);
      color.insert(1.0,1.0,1.0);
    }
  inline void print()
    {
      cout << "PointLight {\n";
      cout << "on "; on.print();
      cout << "location "; location.print();
      cout << "radius "; radius.print();
      cout << "intensity "; intensity.print();
      cout << "color "; color.print();
      cout << "}\n";

    }



};



class transform {

 private:
  vector<shape> shape_list;
  vector<transform*> t_list;
  vector<billboard> b_list;
  vector<pointlight> p_list;


 public:
  SFColor translation;
  SFRotation rotation;
  SFColor scale;
  transform()
    {
      translation.insert(0.0,0.0,0.0);
      rotation.insert(0.0,0.0,1.0,0.0);
      scale.insert(1.0,1.0,1.0);
    }


  insert(shape shape_object)
    {
      shape_list.push_back(shape_object);
    }
  
  insert(transform t_object)
    {
      transform* transform_ptr = new transform(t_object);

      t_list.push_back(transform_ptr);
    }

  insert(billboard b_object)
    {
      b_list.push_back(b_object);
    }
  insert(pointlight p_object)
    {
      p_list.push_back(p_object);
    }
  
  
 inline void print()
   {
     cout << "Transform {\n";
     cout << "translation "; translation.print();
     cout << "rotation "; rotation.print();
     cout << "scale "; scale.print();

     cout << "children [\n";
     for(int i=0;i<shape_list.size();i++)
       {
	 shape_list[i].print();
	 cout << ",\n";
       }
     for(int i=0;i<t_list.size();i++)
       {
	 t_list[i]->print();
	 cout << ",\n";
       }
     for(int i=0;i<b_list.size();i++)
       {
	 b_list[i].print();
	 cout << ",\n";
       }
     for(int i=0;i<p_list.size();i++)
       {
	 p_list[i].print();
	 cout << ",\n";
       }

     cout << "]\n";
     cout << "}\n";
   }
  inline void clear()
    {
      shape_list.erase(shape_list.begin(),shape_list.end());
    }

};



class group {


 public:
  vector<transform> tlist;
  vector<shape> slist;


  inline void insert(transform tnode)
    {
      tlist.push_back(tnode);
    }
  inline void insert(shape snode)
    {
      slist.push_back(snode);
    }
  inline void print()
    {
      int i;
      
      cout << "Group {\n";
      cout << "children [\n";
      for(i=0;i<tlist.size();i++)
	{
	  tlist[i].print();
	  cout << ",\n";
	}
      for(i=0;i<slist.size();i++)
	{
	  slist[i].print();
	  cout << ",\n";
	}
      cout << "]\n";
      cout << "}\n";
    }

};


#endif /* __VRML_H_ */








