#include "vrmlgen.h"
#include "vrml.h"
#include <signal.h>
#include <unistd.h>

/************************************************************

Copyright (C) 1997 Carl Shimer


This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.


 Notes:

This program generates VRML code from the altavista livetopics search
query refinement utility.  Note that if the format of the livetopics
queries changes, this program will crash.

This program should also clean up after itself.  If it is terminated
abnormally, (i.e., the web server kills it), the signal handler will
be called which will call the destructor


This program uses heavy use of the STL template libraries, especially
the <vector> classes.  It has been noted that the robustness and
quality of implementation varies between compiliers.  For instance,
this code compiles and runs perfectly under Digital Unix but has
compilation problems under Linux.

VRML references:

see http://www.vrml.org
 http://vrml.sgi.com

A great book on VRML is the VRML sourcebook by Ames, Nadeau, and
Moreland.



************************************************************/

// global object
internet_handler nethandler;


/************************************************************
get_socket()

This function simply performs standard socket operations 
in order to open a socket to the altavista site.


************************************************************/




int internet_handler::get_socket()
{
  char dest[40] = DEST_HOST;
  struct hostent* host;
  struct sockaddr_in name;
  int soc;

  bzero( (void*)&name,sizeof(name));
  name.sin_family = AF_INET;  /* Internet Protocol */

  /* Get the hosts Internet address */
  if ((host = gethostbyname (dest)) == NULL)
    {
      printf("host unknown, exiting\n");
      return -1;
    }
  
  bcopy( host->h_addr, (void*)&name.sin_addr, host->h_length);

      /* Port to connect to. Note you need to convert from Host long to */
      /* network long representation */
      
      /*  name.sin_port = htons(cport);*/
      name.sin_port = htons(DEST_PORT);
      
      /* Create a new Stream socket in Internet domain */
      if ((soc = socket( AF_INET, SOCK_STREAM, IP)) < 0)
      {
        fprintf(stderr,"Cannot open socket\n");
        return -1;
      }
        
    /* Connect this socket to the host and port specified in command line */
    if ((connect(soc,(struct sockaddr*)&name,sizeof(name))) < 0)
      {
        fprintf(stderr,"Cannot connect to server\n");
        return -1;
      }

return soc;
}

/************************************************************

altavista_io(char*)

This function performs all IO with the altavista http server.  First,
it generates the request string for altavista.  Next, it opens up a
socket to altavista.  It writes out then request and reads until the
socket is closed.  The function reads all data into a temporary file.

************************************************************/


int internet_handler::altavista_io(char* input)
{
  int socket;
  char request[200];
  int size;
  int number=0;
  char* tstring,*oldstring;
  char buffer[1000];

  // open file for writing

  ofstream output(filename);

  if(!output)
    {
      cout << "cannot open file " << filename << endl;
      return -1;
    }

  // assemble our input string

  strcpy(request,BASE_STRING);
  
  tstring = strtok(input,"+");
  if(tstring != NULL)
    {
      strcat(request,tstring);
      while((tstring = strtok(NULL,"+")) != NULL)
	{
	  number++;
	  oldstring = tstring;
	  strcat(request,"+");
	  strcat(request,tstring);

	}

    }
  
  if(number > 0)
    {
      strcat(request,"&r=");
      strcat(request,oldstring);
      strcat(request,"+");

    }


  if((socket = get_socket())<0)
    {
      fprintf(stderr,"get_socket failed\n");
      return -1;
    }
  
  
  write(socket,request,strlen(request));
  
  
  while((size = read(socket,buffer,1000)) > 0)
    {
     output.write(buffer,size);

    }

  output.close();
  close(socket);
    


  }


/************************************************************

topic::insert()

This function parses an input string and stores the data in the
relevent data structures inside a topic class.  See the topic class in
the header files for more information.

************************************************************/


void topic::insert(char* list)
{
char buffer[50];
char* c = list;
int i;
int priority;
int size = (int)list + strlen(list);
searchpair tpair;

index = 0;

while(*c != '\0' && (int)c <= size) // while not end
  {
    bzero(buffer,50);               // clear buffer
    
    for(i=0;*c!=' ';i++,c++)        // store string
      buffer[i] = *c;
    buffer[i] = '\0';
    c++;
    priority = atoi(c++);          // store priority
    
   c++;
 
    tpair.insert(buffer,priority);
    
    insert_element(tpair);
  }
}

/************************************************************

internet_handler::extract_into()

This code parses the HTML file from altavista and extracts the
relevent data and stores it in a list of topics and links.

************************************************************/


void internet_handler::extract_info()
{
char* buffer = new char[1000];
char tbuff[10];
char* tptr;
ifstream fileptr(filename);
topic newtopic;
int link1,link2;
int i;

if(!fileptr)
  {
    cout << "cannot open file " << filename << endl;
    return;
  }


while(!fileptr.eof())
  {
    fileptr.gets(&buffer);
    if(strncasecmp(buffer,TOPIC_SEARCH,strlen(TOPIC_SEARCH)) == 0)
       {

	 // strip off tailing junk
	 buffer[strlen(buffer)] = '\0';
	 buffer[strlen(buffer)-1] = '\0';

	 tptr = buffer;
	 tptr += strlen(TOPIC_SEARCH);
	 
	 
	 tptr = strtok(tptr,";");

	 newtopic.insert(tptr);
	 insert_topic(newtopic);

	 while((tptr = strtok(NULL,";")) != NULL)
	   {
	     newtopic.insert(tptr);
	     insert_topic(newtopic);
	   }
       }
    else if(strncasecmp(buffer,TOPIC_LINK,strlen(TOPIC_LINK)) == 0)
      {
	// clear out extraneous junk

	buffer[strlen(buffer)] = '\0';
	buffer[strlen(buffer)-1] = '\0';

	tptr = buffer;
	tptr += strlen(TOPIC_SEARCH);
	
	while(*tptr != '\0' && *tptr != '\"')   //"
	  {
	    bzero(tbuff,10);

	    for(i=0;*tptr != '-';i++,tptr++)
	      tbuff[i] = *tptr;
	    tbuff[i] = '\0';
	    tptr++;
	    link1 = atoi(tbuff);

	    for(i=0;*tptr != ',' && *tptr != '\0';i++,tptr++)
	      tbuff[i] = *tptr;
	    tbuff[i] = '\0';
	    tptr++;
	    link2 = atoi(tbuff);
	    
	    linkrelations tlink(link1,link2);
	    insert_link(tlink);
	    
	  }
      }
  }
fileptr.close();
}

/************************************************************

internet_handler::gen_cone()

This function generates the cone part of a cone tree.  It first
generates the root node and draws all of the children as the base of
the cone.  This function also adds line between the root node and the
children.

This function retruns a transform object.  This object is then
inserted into either another transform object or into the main object.

************************************************************/


transform* internet_handler::gen_cone(topic current_topic)
{
shape line_shape,*base_shape,shape_object;
box box_object;
cone cone_object;
geom_container<box,cone,text,IndexedLineSet> newcontainer,base_container;
float x,y,z,angle,rad_angle;
int radius,i;
billboard billboard_object;
transform* full_cone,*tnode;
IndexedLineSet lineset;
SFColor linedimension;
pointlight p_object;

full_cone = new transform;
float color[3];

// first, place first node at the top and don't do any translation


newcontainer.insert(cone_object);
shape_object.insert(newcontainer);
// color the top node based on priority
// float color = (float)current_topic.topiclist[0].get_priority()/(float)10;

color[0] = color_table[current_topic.topiclist[0].get_priority()-1][0];
color[1] = color_table[current_topic.topiclist[0].get_priority()-1][1];
color[2] = color_table[current_topic.topiclist[0].get_priority()-1][2];

shape_object.appearance_node.material_node.emissiveColor.insert(color[0],
								color[1],
								color[2]);
//shape_object.appearance_node.material_node.diffuseColor.insert(1,0,0);

p_object.radius.insert(5);
p_object.intensity.insert((float)current_topic.topiclist[0].get_priority()/10);
full_cone->insert(p_object);


full_cone->insert(shape_object);

// next insert text of our object

/*
temp_shape = new shape;
text_object.insert(current_topic.return_name(0));
geom_container<box,cone,text,IndexedLineSet> t_container;
t_container.insert(text_object);
temp_shape->insert(t_container);
billboard_object.insert(*temp_shape);
tnode = new transform;
tnode->insert(billboard_object);
tnode->translation.insert(0,1,0);
full_cone->insert(*tnode);
*/

// setup our line shape
linedimension.insert(0,0,0);
lineset.coord_object.insert(linedimension);

// set the variable sizes

radius = 5;
y = -10;
x = radius;
z = 0;





for(i=0,angle=360/current_topic.size();i<current_topic.size();i++,
    angle+=360/current_topic.size())
  {
    base_container.insert(box_object);
    base_shape = new shape;
    base_shape->insert(base_container);
    base_shape->appearance_node.material_node.emissiveColor.insert(
color_table[current_topic.topiclist[i].get_priority()-1][0],color_table[current_topic.topiclist[i].get_priority()-1][1],color_table[current_topic.topiclist[i].get_priority()-1][2]);
								  

    tnode = new transform;
    tnode->insert(*base_shape);
    tnode->translation.insert(x,y,z);
    full_cone->insert(*tnode);

    linedimension.insert(x,y,z);
    lineset.coord_object.insert(linedimension);
    lineset.index_object.insert(0,i);

    rad_angle = angle*PI/180;
    x = cos(rad_angle)*radius;
    z = sin(rad_angle)*radius;

  }
linedimension.insert(x,y,z);
lineset.coord_object.insert(linedimension);
lineset.index_object.insert(0,i);

geom_container<box,cone,text,IndexedLineSet> linecontain;
linecontain.insert(lineset);
line_shape.insert(linecontain);
line_shape.appearance_node.material_node.emissiveColor.insert(1,1,0);
full_cone->insert(line_shape);

return full_cone;



}

/************************************************************

internet_handler::get_sub_links(vector<int>&,int)

This function parses the vector that contains the list of links
between nodes.  If there is a link to the given node, it pushes that
link into a new vector.  It also sets flags so the program knows not
to look at this link again.


************************************************************/

void internet_handler::get_sub_links(vector<int>& cone,int node)
{
for(int i=0;i<link_size();i++)
  {
    if(available_links[i].relation1 == node)
      {
	if(available_topics[available_links[i].relation2].validity())
	  {
	    available_topics[available_links[i].relation2].invalidate();
	      cone.push_back(available_links[i].relation2);
	  }

      }
    else if(available_links[i].relation2 == node)
      {
	if(available_topics[available_links[i].relation1].validity())
	  {
	    available_topics[available_links[i].relation1].invalidate();
	    cone.push_back(available_links[i].relation1);
	  }
      }
  }
}

/************************************************************

get_peer_links(vector<int>&,int,int)


This function is very similar to get_sub_links except that it looks if
they both have the same level of priority.  Priority determines the
relevence of the query and if both priorities are equal they are
considered peers.

************************************************************/


void internet_handler::get_peer_links(vector<int>& cone,int node,int priority)
{
  for(int i=0;i<link_size();i++)
  {
    if(available_links[i].relation1 == node)
      {
	if(available_topics[available_links[i].relation2].validity())
	  {
	    if(available_topics[available_links[i].relation2].topiclist[0].get_priority()
	       == priority)
	      {
		
		available_topics[available_links[i].relation2].invalidate();
		cone.push_back(available_links[i].relation2);
	      }
	  }

      }
    else if(available_links[i].relation2 == node)
      {
	if(available_topics[available_links[i].relation1].validity())
	  {
	     if(available_topics[available_links[i].relation1].topiclist[0].get_priority()
		== priority)
	      {

		available_topics[available_links[i].relation1].invalidate();
		cone.push_back(available_links[i].relation1);
	      }
	   }
      }
  }
}


transform* internet_handler::
gen_level_cone(SFColor coords,SFColor top_coords,int radius,vector<int> large_cone)
{
IndexedLineSet lineset;  
SFColor linedimension;
float x,y,z;
float angle,rad_angle;
transform* transform_ptr,* group_object;
shape line_shape;
int i;

x = coords.s1;
y = coords.s2;
z = coords.s3;

// alloc memory
transform_ptr = new transform;
group_object = new transform;

// setup lines
linedimension.insert(top_coords.s1,top_coords.s2,top_coords.s3);
lineset.coord_object.insert(linedimension);

if(large_cone.size() != 0)
  {
    
    for(i=0,angle=360/large_cone.size();i<large_cone.size();i++,angle+=360/large_cone.size())
      {
	transform_ptr = gen_cone(available_topics[large_cone[i]]);
	transform_ptr->translation.insert(x,y,z);
	available_topics[large_cone[i]].insert_coords(x,y,z);
      
	linedimension.insert(x,y,z);
	lineset.coord_object.insert(linedimension);
	lineset.index_object.insert(0,i);
      
	rad_angle = angle*PI/180;
	x = cos(rad_angle)*radius;
	z = sin(rad_angle)*radius;
	
	
	group_object->insert(*transform_ptr);
      }
    linedimension.insert(x,y,z);
    lineset.coord_object.insert(linedimension);
    lineset.index_object.insert(0,i);
    
    
    geom_container<box,cone,text,IndexedLineSet> linecontain;
    linecontain.insert(lineset);
    line_shape.insert(linecontain);
    line_shape.appearance_node.material_node.emissiveColor.insert(1,1,0);
    group_object->insert(line_shape);
    
    return group_object;
  }
return NULL;
}

float check_add(float old_value,float new_value)
{
  if(old_value < 0)
    return old_value - new_value;
  else
    return old_value + new_value;


}


transform* internet_handler::draw_peers(vector<int> peers,int node,int radius)
{
  IndexedLineSet lineset;  
  SFColor linedimension;
  float angle;
  float x,y,z;
  transform* new_transform,* t_ptr;
  int i=0;
  float new_x, new_z;
  shape line_shape;
  vector<int> new_peers;

  new_transform = new transform;

  if(peers.size() == 0)
    return NULL;


  // artan(z,x)
  
  if(peers.size() > i)
    {
      linedimension.insert(available_topics[node].coords[0],
			   available_topics[node].coords[1],
			   available_topics[node].coords[2]);

      
      lineset.coord_object.insert(linedimension);
      
      angle = atan(available_topics[node].coords[2]/available_topics[node].coords[0]);
      
      
      radius += 10;
      z = sin(angle)*radius;
      new_z = z;
      z = check_add(available_topics[node].coords[2],z);
      x = cos(angle)*radius;
      new_x = x;
      x = check_add(available_topics[node].coords[0],x);
      y = available_topics[node].coords[1];
      
      // set our coordinates! very important

      available_topics[peers[0]].insert_coords(x,y,z);
      
      t_ptr = gen_cone(available_topics[peers[0]]);
      t_ptr->translation.insert(x,y,z);
      new_transform->insert(*t_ptr);
  


    
      linedimension.insert(x,y,z);
      lineset.coord_object.insert(linedimension);
      lineset.index_object.insert(0,i+1);

      new_peers.erase(new_peers.begin(),new_peers.end());

      get_peer_links(new_peers,peers[0],
       available_topics[peers[0]].topiclist[0].get_priority());
      

      t_ptr = draw_peers(new_peers,peers[0],radius);
      if(t_ptr != NULL)
	new_transform->insert(*t_ptr);

    }
  i++;
  
  if(peers.size() > i)
    {
      x = check_add(new_x,x);
      z = available_topics[node].coords[2];
      y = available_topics[node].coords[1];

      available_topics[peers[1]].insert_coords(x,y,z);

      t_ptr = gen_cone(available_topics[peers[1]]);
      t_ptr->translation.insert(x,y,z);
      new_transform->insert(*t_ptr);
      
      linedimension.insert(x,y,z);
      lineset.coord_object.insert(linedimension);
      lineset.index_object.insert(0,i+1);

      new_peers.erase(new_peers.begin(),new_peers.end());
      get_peer_links(new_peers,peers[1],
		     available_topics[peers[1]].topiclist[0].get_priority());
      t_ptr = draw_peers(new_peers,peers[1],radius);
      if(t_ptr != NULL)
	new_transform->insert(*t_ptr);
      
    }
  i++;
  if(peers.size() > i)
    {
      x = available_topics[node].coords[0];
      z = check_add(new_z,z);
      y = available_topics[node].coords[1];
      
      available_topics[peers[2]].insert_coords(x,y,z);

      t_ptr = gen_cone(available_topics[peers[2]]);
      t_ptr->translation.insert(x,y,z);
      new_transform->insert(*t_ptr);

      linedimension.insert(x,y,z);
      lineset.coord_object.insert(linedimension);
      lineset.index_object.insert(0,i+1);

      new_peers.erase(new_peers.begin(),new_peers.end());
      get_peer_links(new_peers,peers[2],
			     available_topics[peers[2]].topiclist[0].get_priority());
      t_ptr = draw_peers(new_peers,peers[2],radius);
      if(t_ptr != NULL)
	new_transform->insert(*t_ptr);

    }
  geom_container<box,cone,text,IndexedLineSet> linecontain;
  linecontain.insert(lineset);
  line_shape.insert(linecontain);
  line_shape.appearance_node.material_node.emissiveColor.insert(1,1,0);
  new_transform->insert(line_shape);

return new_transform;
}


/************************************************************

produce webpage: major calls to VRML library

************************************************************/

int internet_handler::find_node()
{
int highest_priority=-1;
int keep_node;
int temp;

for(int i=0;i<topic_size();i++)
  {
    if(available_topics[i].validity())
      if((temp = available_topics[i].topiclist[0].get_priority())
	 > highest_priority)
	{
	  highest_priority = temp;
	  keep_node = i;
	}
  }

if(highest_priority == -1)
  return -1;
else return keep_node;


}


int internet_handler::produce_webpage()
{

transform transform_object,*transform_ptr;
transform* group_object;
group total_object;
vector<int> large_cone;
vector<int> peers;

int i;
float x,y,z,angle;
int radius;
int current_node = 0;
SFColor coords,top_coords;
SFColor top_translate;
int j;
int top_node;


coords.insert(0,0,0);
top_translate.insert(0,0,0);
available_topics[0].insert_coords(0,0,0);



top_coords.insert(0,0,0);

// initial offset values
radius = 15;
y = -20;
x = radius;
z = 0;
    coords.insert(x,y,z);




while((top_node = find_node()) != -1)
  {
    fprintf(stderr,"top_node is %d\n",top_node);
    group_object = new transform;

    group_object->insert(*gen_cone(available_topics[top_node]));
    
    // this is not really the right place to be doing this 
    available_topics[top_node].insert_coords(0,0,0);

    get_sub_links(large_cone,top_node);
    current_node = top_node;
    available_topics[top_node].invalidate();
    
    while(large_cone.size() != 0 && current_node >= 0)
      {
	available_topics[current_node].invalidate();
	
	transform_ptr = gen_level_cone(coords,top_coords,radius,large_cone);
	if(transform_ptr != NULL)
	  group_object->insert(*transform_ptr);
	
	available_topics[current_node].invalidate_children();
	
	
	for(i=0;i<large_cone.size();i++)
	  {
	    peers.erase(peers.begin(),peers.end());
	    get_peer_links(peers,large_cone[i],
			   available_topics[large_cone[i]].topiclist[0].get_priority());
	    
	    
	    
	    transform_ptr = draw_peers(peers,large_cone[i],radius);
	    if(transform_ptr != NULL)
	      group_object->insert(*transform_ptr);
	  }
	
	
	current_node = -1;
	for(j=0;j<topic_size();j++)
	  if(!available_topics[j].validity() && 
	     !available_topics[j].drawn_children_query())
	    {
	      current_node = j;
	      break;
	    }
	
	
	large_cone.erase(large_cone.begin(),large_cone.end());
	get_sub_links(large_cone,current_node);
	
	top_coords.insert(available_topics[current_node].coords[0],
			  available_topics[current_node].coords[1],
			  available_topics[current_node].coords[2]);
	coords.insert(top_coords.s1 - radius,
		      top_coords.s2 - 20,
		      top_coords.s3);
	
    }
    group_object->translation.insert(top_translate.s1,
				    top_translate.s2,
				    top_translate.s3);
    total_object.insert(*group_object);
  
  }  
float rad_angle;
   
for(i=1,angle = 360/total_object.tlist.size()-1;
    i<total_object.tlist.size();i++,
    angle+=360/total_object.tlist.size())
  {
    radius = 30;

    rad_angle = angle*PI/180;
    x = cos(rad_angle)*radius;
    z = sin(rad_angle)*radius;
    y = 0;

    total_object.tlist[i].translation.insert(x,y,z);
  }

  
// generate the VRML
total_object.print();
}


void signal_handler(int unused)
{
// force object to clean up

nethandler.~internet_handler();
exit(1);

}


int main()
{
char* cl;
signal(SIGINT,signal_handler);
signal(SIGQUIT,signal_handler);
signal(SIGTERM,signal_handler);

cout << "Content-type: model/vrml\n\n";


cl = getenv("QUERY_STRING");
strtok(cl,"=");

nethandler.altavista_io(strtok(NULL,"\0"));
nethandler.extract_info();

// insert VRML header
cout << VRML_HEADER;


nethandler.produce_webpage();


}












